@@ -6,7 +6,9 @@
 import os
 import re
 import html
+import json
 import traceback
+import unicodedata
 import pkg_resources
 import lupa
 from lupa import LuaRuntime
@@ -25,11 +27,6 @@ builtin_lua_search_paths = [
 lua_dir = pkg_resources.resource_filename("wikitextprocessor", "lua/")
 #print("lua_dir", lua_dir)
 
-# Set of known language codes.
-# XXX remove this?
-#KNOWN_LANGUAGE_TAGS = set(x["code"] for x in ALL_LANGUAGES
-#                          if x.get("code") and x.get("name"))
-
 # Mapping from language code code to language name.
 LANGUAGE_CODE_TO_NAME = { x["code"]: x["name"]
                           for x in ALL_LANGUAGES
@@ -40,14 +37,19 @@ def lua_loader(ctx, modname):
     This will load it from either the user-defined modules on special
     pages or from a built-in module in the file system.  This returns None
     if the module could not be loaded."""
+    # print("lua_loader", modname)
     assert isinstance(modname, str)
+    # XXX consider implementing some kind of cache here, e.g., LRU cache
+    # for the most recently used modules (modname -> text)
     # print("Loading", modname)
     modname = modname.strip()
     if modname.startswith("Module:"):
         modname = modname[7:]
-    modname1 = ctx._canonicalize_template_name(modname)
-    if modname1 in ctx.modules:
-        return ctx.modules[modname1]
+    modname1 = "Module:" + modname
+    modname1 = ctx._canonicalize_template_name(modname1)
+    body = ctx.read_by_title(modname1)
+    if body is not None:
+        return body
     path = modname
     path = re.sub(r":", "/", path)
     path = re.sub(r" ", "_", path)
@@ -65,8 +67,11 @@ def lua_loader(ctx, modname):
             with open(p, "r") as f:
                 data = f.read()
             return data
-    ctx.error("Lua module not found: NOT FOUND: {} at {}"
-              .format(modname, ctx.expand_stack))
+    # It is very common that modules are not found; e.g., many Chinese macros
+    # attempt to load modules related to the characters and generate tons
+    # of these errors.
+    #ctx.error("Lua module not found: {}"
+    #          .format(modname))
     return None
 
 
@@ -112,6 +117,67 @@ def mw_text_encode(text, charset='<>&\xa0"'):
             parts.append(ch)
     return "".join(parts)
 
+def mw_text_jsondecode(ctx, s, *rest):
+    flags = rest[0] if rest else 0
+    value = json.loads(s)
+
+    def recurse(x):
+        if isinstance(x, (list, tuple)):
+            return ctx.lua.table_from(list(map(recurse, x)))
+        if not isinstance(x, dict):
+            return x
+        # It is a dict.
+        if (flags & 1) == 1:
+            # JSON_PRESERVE_KEYS flag means we don't convert keys.
+            return ctx.lua.table_from({k: recurse(v) for k, v in x.items()})
+        # Convert numeric keys to integers and see if we can make it a
+        # table with sequential integer keys.
+        for k, v in list(x.items()):
+            if k.isdigit():
+                del x[k]
+                x[int(k)] = recurse(v)
+            else:
+                x[k] = recurse(v)
+        if not all(isinstance(k, int) for k in x.keys()):
+            return ctx.lua.table_from(x)
+        keys = list(sorted(x.keys()))
+        if not all(keys[i] == i + 1 for i in range(len(keys))):
+            return ctx.lua.table_from(x)
+        values = list(x[i + 1] for i in range(len(keys)))
+        return ctx.lua.table_from(x)
+
+    value = recurse(value)
+    return value
+
+def mw_text_jsonencode(s, *rest):
+    flags = rest[0] if rest else 0
+
+    def recurse(x):
+        if isinstance(x, (str, int, float, type(None), type(True))):
+            return x
+        if lupa.lua_type(x) == "table":
+            conv_to_dict = (flags & 1) != 0  # JSON_PRESERVE_KEYS flag
+            if not conv_to_dict:
+                # Also convert to dict if keys are not sequential integers
+                # starting from 1
+                if not all(isinstance(k, int) for k in x.keys()):
+                    conv_to_dict = True
+                else:
+                    keys = list(sorted(x.keys()))
+                    if not all(keys[i] == i + 1 for i in range(len(keys))):
+                        conv_to_dict = True
+            if conv_to_dict:
+                ht = {}
+                for k, v in x.items():
+                    ht[str(k)] = recurse(v)
+                return ht
+            # Convert to list (JSON array)
+            return list(map(recurse, x.values()))
+        return x
+
+    value = recurse(s)
+    return json.dumps(value, sort_keys=True)
+
 
 def get_page_info(ctx, title):
     """Retrieves information about a page identified by its table (with
@@ -145,10 +211,10 @@ def get_page_content(ctx, title):
     # Read the page by its title
     data = ctx.read_by_title(title)
     if data is None:
-        ctx.warning("attempted to access page content for {!r} which "
-                    "is not available"
-                    .format(title))
-        return ""
+        # ctx.warning("attempted to access page content for {!r} which "
+        #             "is not available"
+        #             .format(title))
+        return None
     return data
 
 def fetch_language_name(code):
@@ -188,6 +254,9 @@ def initialize_lua(ctx):
     lua.eval("lua_set_loader")(lambda x: lua_loader(ctx, x),
                                mw_text_decode,
                                mw_text_encode,
+                               mw_text_jsonencode,
+                               lambda x, *rest:
+                               mw_text_jsondecode(ctx, x, *rest),
                                lambda x: get_page_info(ctx, x),
                                lambda x: get_page_content(ctx, x),
                                fetch_language_name,
@@ -195,23 +264,22 @@ def initialize_lua(ctx):
     ctx.lua = lua
 
 
-def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
+def call_lua_sandbox(ctx, invoke_args, expander, parent):
     """Calls a function in a Lua module in the Lua sandbox.  This creates
     the sandbox instance if it does not already exist and stores it in
     ctx.lua.  ``invoke_args`` is the arguments to the call;
-    ``expander`` should be a function to expand an argument.  Stack
-    indicates how we ended up in this call; it is for debugging and
-    error messages only.  This will restore stack to as it was before
-    returning.  ``parent`` should be None or (parent_title,
-    parent_args) for the parent page."""
+    ``expander`` should be a function to expand an argument.
+    ``parent`` should be None or (parent_title, parent_args) for the
+    parent page."""
     assert isinstance(invoke_args, (list, tuple))
     assert callable(expander)
-    assert isinstance(stack, list)
     assert parent is None or isinstance(parent, (list, tuple))
 
+    # print("call_lua_sandbox", invoke_args, parent)
+
     if len(invoke_args) < 2:
-        ctx.error("#invoke {}: too few arguments at {}"
-                  "".format(invoke_args, stack))
+        ctx.error("#invoke {}: too few arguments"
+                  .format(invoke_args))
         return ("{{" + invoke_args[0] + ":" +
                 "|".join(invoke_args[1:]) + "}}")
 
@@ -236,6 +304,7 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
         assert isinstance(args, (list, tuple, dict))
         # Convert args to a dictionary with default value None
         if isinstance(args, dict):
+            args = {k: html.unescape(v) for k, v in args.items()}
             frame_args = args
         else:
             assert isinstance(args, (list, tuple))
@@ -257,13 +326,13 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
                     # No argument name
                     k = num
                     num += 1
+                arg = html.unescape(arg)
                 frame_args[k] = arg
         frame_args = lua.table_from(frame_args)
 
         def extensionTag(frame, *args):
             if len(args) < 1:
-                ctx.error("extensionTag: missing arguments at {}"
-                          .format(stack))
+                ctx.error("extensionTag: missing arguments")
                 return ""
             dt = args[0]
             if not isinstance(dt, (str, int, float, type(None))):
@@ -294,19 +363,17 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
             else:
                 attrs = [attrs]
 
-            stack.append("extensionTag()")
+            ctx.expand_stack.append("extensionTag()")
             ret = tag_fn(title, "#tag", [name, content] + attrs,
-                         lambda x: x,  # Already expanded
-                         stack)
-            stack.pop()
+                         lambda x: x)  # Already expanded
+            ctx.expand_stack.pop()
             # Expand any templates from the result
             ret = preprocess(frame, ret)
             return ret
 
         def callParserFunction(frame, *args):
             if len(args) < 1:
-                ctx.error("callParserFunction: missing name at {}"
-                          .format(stack))
+                ctx.error("callParserFunction: missing name")
                 return ""
             name = args[0]
             if not isinstance(name, str):
@@ -328,23 +395,22 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
                         new_args.append(str(v))
             name = ctx._canonicalize_parserfn_name(name)
             if name not in PARSER_FUNCTIONS:
-                ctx.error("frame:callParserFunction(): undefined function "
-                          "{!r} at {}".format(name, stack))
+                ctx.error("frame:callParserFunction(): undefined function {!r}"
+                          .format(name))
                 return ""
-            return call_parser_function(ctx, name, new_args, lambda x: x,
-                                        stack)
+            return call_parser_function(ctx, name, new_args, lambda x: x)
 
         def expand_all_templates(encoded):
             # Expand all templates here, even if otherwise only
             # expanding some of them.  We stay quiet about undefined
             # templates here, because Wiktionary Module:ugly hacks
             # generates them all the time.
-            ret = ctx.expand(encoded, stack, parent, quiet=True)
+            ret = ctx.expand(encoded, parent, quiet=True)
             return ret
 
         def preprocess(frame, *args):
             if len(args) < 1:
-                ctx.error("preprocess: missing argument at {}".format(stack))
+                ctx.error("preprocess: missing argument")
                 return ""
             v = args[0]
             if not isinstance(v, str):
@@ -352,20 +418,18 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
             # Expand all templates, in case the Lua code actually
             # inspects the output.
             v = ctx._encode(v)
-            stack.append("frame:preprocess()")
+            ctx.expand_stack.append("frame:preprocess()")
             ret = expand_all_templates(v)
-            stack.pop()
+            ctx.expand_stack.pop()
             return ret
 
         def expandTemplate(frame, *args):
             if len(args) < 1:
-                ctx.error("expandTemplate: missing arguments at {}"
-                          "".format(stack))
+                ctx.error("expandTemplate: missing arguments")
                 return ""
             dt = args[0]
             if isinstance(dt, (int, float, str, type(None))):
-                ctx.error("expandTemplate: arguments should be named at {}"
-                          "".format(stack))
+                ctx.error("expandTemplate: arguments should be named")
                 return ""
             title = dt["title"] or ""
             args = dt["args"] or {}
@@ -373,9 +437,9 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
             for k, v in sorted(args.items(), key=lambda x: str(x[0])):
                 new_args.append("{}={}".format(k, v))
             encoded = ctx._save_value("T", new_args, False)
-            stack.append("frame:expandTemplate()")
+            ctx.expand_stack.append("frame:expandTemplate()")
             ret = expand_all_templates(encoded)
-            stack.pop()
+            ctx.expand_stack.pop()
             return ret
 
         # Create frame object as dictionary with default value None
@@ -413,9 +477,8 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
     frame = make_frame(pframe, modname, invoke_args[2:])
 
     # Call the Lua function in the given module
-    stack.append("Lua:{}:{}()".format(modname, modfn))
-    old_stack = ctx.expand_stack
-    ctx.expand_stack = stack
+    stack_len = len(ctx.expand_stack)
+    ctx.expand_stack.append("Lua:{}:{}()".format(modname, modfn))
     try:
         ret = lua.eval("lua_invoke")(modname, modfn, frame, ctx.title)
         if not isinstance(ret, (list, tuple)):
@@ -425,16 +488,18 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
         else:
             ok, text = ret[0], ret[1]
     except UnicodeDecodeError:
-        ctx.error("ERROR: {}: invalid unicode returned by {} at {}"
-                  .format(ctx.title, invoke_args, stack))
+        ctx.error("invalid unicode returned by {} parent {}"
+                  .format(invoke_args, parent))
         ok, text = True, ""
     finally:
-        ctx.expand_stack = old_stack
-    stack.pop()
+        while len(ctx.expand_stack) > stack_len:
+            ctx.expand_stack.pop()
     if ok:
         if text is None:
             text = "nil"
-        return str(text)
+        text = str(text)
+        text = unicodedata.normalize("NFC", text)
+        return text
     if isinstance(text, Exception):
         parts = [str(text)]
         lst = traceback.format_exception(etype=type(text),
@@ -445,8 +510,15 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
         text = "\n".join(parts)
     elif not isinstance(text, str):
         text = str(text)
+    msg = re.sub(r".*?:\d+: ", "", text.split("\n")[0])
     if text.find("'debug.error'") >= 0:
-        ctx.warning(re.sub(r".*?:\d+: ", "", text.split("\n")[0]))
+        if not msg.startswith("This template is deprecated."):
+            ctx.warning(msg)
+    # XXX This probably should be deleted; it can cause important errors
+    # to be missed
+    # elif text.find("in function 'error'") >= 0:
+    #     if not msg.startswith('The parameter "'):
+    #         ctx.warning(msg)
     else:
         parts = []
         in_traceback = 0
@@ -456,7 +528,11 @@ def call_lua_sandbox(ctx, invoke_args, expander, stack, parent):
             #    break
             parts.append(line)
         trace = "\n".join(parts)
-        ctx.error("LUA error in #invoke {} at {}"
-                  .format(invoke_args, stack),
+        ctx.error("LUA error in #invoke {} parent {}"
+                  .format(invoke_args, parent),
                   trace=trace)
-    return ""
+    msg = "Lua execution error"
+    if text.find("Lua timeout error") >= 0:
+        msg = "Lua timeout error"
+    return ('<strong class="error">{} in Module:{} function {}'
+            '</strong>'.format(msg, html.escape(modname), html.escape(modfn)))
