@@ -9,11 +9,38 @@ import datetime
 import urllib.parse
 import dateparser
 from .wikihtml import ALLOWED_HTML_TAGS
-
+from .common import preprocess_text
 
 # Name of the WikiMedia for which we are generating content
 PROJECT_NAME = "Wiktionary"
 
+# The host to which generated URLs will point
+SERVER_NAME = "dummy.host"
+
+namespace_prefixes = set([
+    "Appendix",
+    "Category",
+    "Citations",
+    "Concordance",
+    "File",
+    "Help",
+    "Image",
+    "Index",
+    "Media",
+    "MediaWiki",
+    "Module",
+    "Project",
+    "Reconstruction",
+    "Rhymes",
+    "Sign gloss",
+    "Summary",
+    "Talk",
+    "Template",
+    "Thesaurus",
+    "Thread",
+    "User",
+    "Wiktionary",
+])
 
 def capitalizeFirstOnly(s):
     if s:
@@ -21,7 +48,7 @@ def capitalizeFirstOnly(s):
     return s
 
 
-def if_fn(ctx, fn_name, args, expander, stack):
+def if_fn(ctx, fn_name, args, expander):
     """Implements #if parser function."""
     arg0 = args[0] if args else ""
     arg1 = args[1] if len(args) >= 2 else ""
@@ -32,7 +59,7 @@ def if_fn(ctx, fn_name, args, expander, stack):
     return expander(arg2).strip()
 
 
-def ifeq_fn(ctx, fn_name, args, expander, stack):
+def ifeq_fn(ctx, fn_name, args, expander):
     """Implements #ifeq parser function."""
     arg0 = args[0] if args else ""
     arg1 = args[1] if len(args) >= 2 else ""
@@ -42,12 +69,27 @@ def ifeq_fn(ctx, fn_name, args, expander, stack):
         return expander(arg2).strip()
     return expander(arg3).strip()
 
-def ifexpr_fn(ctx, fn_name, args, expander, stack):
+
+def iferror_fn(ctx, fn_name, args, expander):
+    """Implements the #iferror parser function."""
+    arg0 = expander(args[0]) if args else ""
+    arg1 = args[1] if len(args) >= 2 else None
+    arg2 = args[2] if len(args) >= 3 else None
+    if re.search(r'<[^>]*?\sclass="error"', arg0):
+        if arg1 is None:
+            return ""
+        return expander(arg1).strip()
+    if arg2 is None:
+        return arg0
+    return expander(arg2).strip()
+
+
+def ifexpr_fn(ctx, fn_name, args, expander):
     """Implements #ifexpr parser function."""
     arg0 = args[0] if args else "0"
     arg1 = args[1] if len(args) >= 2 else ""
     arg2 = args[2] if len(args) >= 3 else ""
-    cond = expr_fn(ctx, fn_name, [arg0], expander, stack)
+    cond = expr_fn(ctx, fn_name, [arg0], expander)
     try:
         ret = int(cond)
     except ValueError:
@@ -56,20 +98,17 @@ def ifexpr_fn(ctx, fn_name, args, expander, stack):
         return expander(arg1)
     return expander(arg2)
 
-def ifexist_fn(ctx, fn_name, args, expander, stack):
+def ifexist_fn(ctx, fn_name, args, expander):
     """Implements #ifexist parser function."""
     arg0 = args[0] if args else ""
     arg1 = args[1] if len(args) >= 2 else ""
     arg2 = args[2] if len(args) >= 3 else ""
-    # XXX how will we check if the page exists?  Should probably pass a context
-    # to all parser function implementations and use the context to evaluate
-    # this.  XXX collect the info in phase1
-    exists = False  # XXX needs to be implemented
+    exists = ctx.page_exists(expander(arg0).strip())
     if exists:
         return expander(arg1).strip()
     return expander(arg2).strip()
 
-def switch_fn(ctx, fn_name, args, expander, stack):
+def switch_fn(ctx, fn_name, args, expander):
     """Implements #switch parser function."""
     val = expander(args[0]).strip() if args else ""
     match_next = False
@@ -77,7 +116,7 @@ def switch_fn(ctx, fn_name, args, expander, stack):
     last = None
     for i in range(1, len(args)):
         arg = expander(args[i]).strip()
-        m = re.match(r"(?s)^([^=]+)=(.*)$", arg)
+        m = re.match(r"(?s)^([^=]*)=(.*)$", arg)
         if not m:
             last = arg
             if arg == val:
@@ -96,15 +135,24 @@ def switch_fn(ctx, fn_name, args, expander, stack):
     return last or ""
 
 
-def lst_fn(ctx, fn_name, args, expander, stack):
+def categorytree_fn(ctx, fn_name, args, expander):
+    """Implements the #categorytree parser function.  This function accepts
+    keyed arguments."""
+    assert isinstance(args, dict)
+    # We don't currently really implement categorytree.  It is just recognized
+    # and silently ignored.
+    return ""
+
+
+def lst_fn(ctx, fn_name, args, expander):
     """Implements the #lst (alias #section etc) parser function."""
     pagetitle = expander(args[0]).strip() if args else ""
     chapter = expander(args[1]).strip() if len(args) >= 2 else ""
     text = ctx.read_by_title(pagetitle)
     if text is None:
         ctx.warning("{} trying to transclude chapter {!r} from non-existent "
-                    "page {!r} at {}"
-                    .format(fn_name, chapter, pagetitle, stack))
+                    "page {!r}"
+                    .format(fn_name, chapter, pagetitle))
         return ""
 
     parts = []
@@ -115,17 +163,17 @@ def lst_fn(ctx, fn_name, args, expander, stack):
                          text):
         parts.append(m.group(1))
     if not parts:
-        ctx.warning("{} could not find chapter {!r} on page {!r} at {}"
-                    .format(fn_name, chapter, pagetitle, stack))
+        ctx.warning("{} could not find chapter {!r} on page {!r}"
+                    .format(fn_name, chapter, pagetitle))
     return "".join(parts)
 
 
-def tag_fn(ctx, fn_name, args, expander, stack):
+def tag_fn(ctx, fn_name, args, expander):
     """Implements #tag parser function."""
     tag = expander(args[0]).lower() if args else ""
-    if tag not in ALLOWED_HTML_TAGS:
-        print("{}: #tag creating non-allowed tag <{}> - omitted at {}"
-              "".format(ctx, tag, stack))
+    if tag not in ALLOWED_HTML_TAGS and tag != "nowiki":
+        ctx.warning("#tag creating non-allowed tag <{}> - omitted"
+                    .format(tag))
         return "{{" + fn_name + ":" + "|".join(args) + "}}"
     content = expander(args[1]) if len(args) >= 2 else ""
     attrs = []
@@ -134,8 +182,8 @@ def tag_fn(ctx, fn_name, args, expander, stack):
             x = expander(x)
             m = re.match(r"""(?s)^([^=<>'"]+)=(.*)$""", x)
             if not m:
-                print("{}: invalid attribute format {!r} missing name at {}"
-                      "".format(ctx, x, stack))
+                ctx.warning("invalid attribute format {!r} missing name"
+                            .format(x))
                 continue
             name, value = m.groups()
             if not value.startswith('"') and not value.startswith("'"):
@@ -146,54 +194,88 @@ def tag_fn(ctx, fn_name, args, expander, stack):
     else:
         attrs = ""
     if not content:
-        return "<{}{} />".format(tag, attrs)
-    content = html.escape(content, quote=False)
-    return "<{}{}>{}</{}>".format(tag, attrs, content, tag)
+        ret = "<{}{} />".format(tag, attrs)
+    else:
+        ret = "<{}{}>{}</{}>".format(tag, attrs, content, tag)
+    if tag == "nowiki":
+        ret = preprocess_text(ret)
+    return ret
 
 
-def fullpagename_fn(ctx, fn_name, args, expander, stack):
+def fullpagename_fn(ctx, fn_name, args, expander):
     """Implements the FULLPAGENAME magic word/parser function."""
     t = expander(args[0]) if args else ctx.title
     t = re.sub(r"\s+", " ", t)
     t = t.strip()
     ofs = t.find(":")
     if ofs == 0:
-        t = capitalizeFirstOnly(t[1:])
+        # t = capitalizeFirstOnly(t[1:])
+        t = t[1:]
     elif ofs > 0:
         ns = capitalizeFirstOnly(t[:ofs])
-        t = capitalizeFirstOnly(t[ofs + 1:])
+        # t = capitalizeFirstOnly(t[ofs + 1:])
+        t = t[ofs + 1:]
         t = ns + ":" + t
-    else:
-        t = capitalizeFirstOnly(t)
+    #else:
+    #    t = capitalizeFirstOnly(t)
     return t
 
 
-def pagename_fn(ctx, fn_name, args, expander, stack):
+def fullpagenamee_fn(ctx, fn_name, args, expander):
+    """Implements the FULLPAGENAMEE magic word/parser function."""
+    t = fullpagename_fn(ctx, fn_name, args, expander)
+    return wikiurlencode(t)
+
+
+def pagenamee_fn(ctx, fn_name, args, expander):
+    """Implements the PAGENAMEE magic word/parser function."""
+    t = pagename_fn(ctx, fn_name, args, expander)
+    return wikiurlencode(t)
+
+
+def rootpagenamee_fn(ctx, fn_name, args, expander):
+    """Implements the ROOTPAGENAMEE magic word/parser function."""
+    t = rootpagename_fn(ctx, fn_name, args, expander)
+    return wikiurlencode(t)
+
+
+def pagename_fn(ctx, fn_name, args, expander):
     """Implements the PAGENAME magic word/parser function."""
     t = expander(args[0]) if args else ctx.title
     t = re.sub(r"\s+", " ", t)
     t = t.strip()
     ofs = t.find(":")
     if ofs >= 0:
-        t = capitalizeFirstOnly(t[ofs + 1:])
-    else:
-        t = capitalizeFirstOnly(t)
+        # t = capitalizeFirstOnly(t[ofs + 1:])
+        t = t[ofs + 1:]
+    #else:
+    #    t = capitalizeFirstOnly(t)
     return t
 
 
-def basepagename_fn(ctx, fn_name, args, expander, stack):
+def basepagename_fn(ctx, fn_name, args, expander):
     """Implements the BASEPAGENAME magic word/parser function."""
     t = expander(args[0]) if args else ctx.title
     t = re.sub(r"\s+", " ", t)
     t = t.strip()
     ofs = t.rfind("/")
     if ofs >= 0:
-        return t[:ofs]
-    else:
-        return pagename_fn(ctx, fn_name, [t], lambda x: x, stack)
+        t = t[:ofs]
+    return pagename_fn(ctx, fn_name, [t], lambda x: x)
 
 
-def subpagename_fn(ctx, fn_name, args, expander, stack):
+def rootpagename_fn(ctx, fn_name, args, expander):
+    """Implements the ROOTPAGENAME magic word/parser function."""
+    t = expander(args[0]) if args else ctx.title
+    t = re.sub(r"\s+", " ", t)
+    t = t.strip()
+    ofs = t.find("/")
+    if ofs >= 0:
+        t = t[:ofs]
+    return pagename_fn(ctx, fn_name, [t], lambda x: x)
+
+
+def subpagename_fn(ctx, fn_name, args, expander):
     """Implements the SUBPAGENAME magic word/parser function."""
     t = expander(args[0]) if args else ctx.title
     t = re.sub(r"\s+", " ", t)
@@ -202,16 +284,29 @@ def subpagename_fn(ctx, fn_name, args, expander, stack):
     if ofs >= 0:
         return t[ofs + 1:]
     else:
-        return pagename_fn(ctx, fn_name, [t], lambda x: x, stack)
+        return pagename_fn(ctx, fn_name, [t], lambda x: x)
 
-def namespacenumber_fn(ctx, fn_name, args, expander, stack):
+
+def talkpagename_fn(ctx, fn_name, args, expander):
+    """Implements the TALKPAGENAME magic word."""
+    ofs = ctx.title.find(":")
+    if ofs < 0:
+        return "Talk:" + ctx.title
+    if ofs >= 0:
+        prefix = ctx.title[:ofs]
+        if prefix not in namespace_prefixes:
+            return "Talk:" + ctx.title
+        return prefix + "_talk:" + ctx.title[ofs + 1:]
+
+
+def namespacenumber_fn(ctx, fn_name, args, expander):
     """Implements the NAMESPACENUMBER magic word/parser function."""
     # XXX currently hard-coded to return the name space number for the Main
     # namespace
     return 0
 
 
-def namespace_fn(ctx, fn_name, args, expander, stack):
+def namespace_fn(ctx, fn_name, args, expander):
     """Implements the NAMESPACE magic word/parser function."""
     t = expander(args[0]) if args else ctx.title
     t = re.sub(r"\s+", " ", t)
@@ -224,51 +319,118 @@ def namespace_fn(ctx, fn_name, args, expander, stack):
         return ns
     return ""
 
-def currentyear_fn(ctx, fn_name, args, expander, stack):
+
+def subjectspace_fn(ctx, fn_name, args, expander):
+    """Implements the SUBJECTSPACE magic word/parser function.  This
+    implementation is very minimal."""
+    t = expander(args[0]) if args else ctx.title
+    for prefix in namespace_prefixes:
+        if t.startswith(prefix + ":"):
+            return prefix
+    return ""
+
+
+def talkspace_fn(ctx, fn_name, args, expander):
+    """Implements the TALKSPACE magic word/parser function.  This
+    implementation is very minimal."""
+    t = expander(args[0]) if args else ctx.title
+    for prefix in namespace_prefixes:
+        if t.startswith(prefix + ":"):
+            return prefix + "_talk"
+    return "Talk"
+
+
+def server_fn(ctx, fn_name, args, expander):
+    """Implements the SERVER magic word."""
+    return "//{}".format(SERVER_NAME)
+
+
+def servername_fn(ctx, fn_name, args, expander):
+    """Implements the SERVERNAME magic word."""
+    return SERVER_NAME
+
+
+def currentyear_fn(ctx, fn_name, args, expander):
     """Implements the CURRENTYEAR magic word."""
     return str(datetime.datetime.utcnow().year)
 
-def currentmonth_fn(ctx, fn_name, args, expander, stack):
+
+def currentmonth_fn(ctx, fn_name, args, expander):
     """Implements the CURRENTMONTH magic word."""
     return "{:02d}".format(datetime.datetime.utcnow().month)
 
-def currentmonth1_fn(ctx, fn_name, args, expander, stack):
+
+def currentmonth1_fn(ctx, fn_name, args, expander):
     """Implements the CURRENTMONTH1 magic word."""
     return "{:d}".format(datetime.datetime.utcnow().month)
 
-def currentday_fn(ctx, fn_name, args, expander, stack):
+
+def currentmonthname_fn(ctx, fn_name, args, expander):
+    """Implements the CURRENTMONTHNAME magic word."""
+    # XXX support for other languages?
+    month = datetime.datetime.utcnow().month
+    return ("", "January", "February", "March", "April", "May", "June",
+            "July", "August", "September", "October", "November",
+            "December")[month]
+
+
+def currentmonthabbrev_fn(ctx, fn_name, args, expander):
+    """Implements the CURRENTMONTHABBREV magic word."""
+    # XXX support for other languages?
+    month = datetime.datetime.utcnow().month
+    return ("", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
+            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")[month]
+
+
+def currentday_fn(ctx, fn_name, args, expander):
     """Implements the CURRENTDAY magic word."""
     return "{:d}".format(datetime.datetime.utcnow().day)
 
-def currentday2_fn(ctx, fn_name, args, expander, stack):
+
+def currentday2_fn(ctx, fn_name, args, expander):
     """Implements the CURRENTDAY2 magic word."""
     return "{:02d}".format(datetime.datetime.utcnow().day)
 
-def currentdow_fn(ctx, fn_name, args, expander, stack):
+
+def currentdow_fn(ctx, fn_name, args, expander):
     """Implements the CURRENTDOW magic word."""
     return "{:d}".format(datetime.datetime.utcnow().weekday())
 
 
-def displaytitle_fn(ctx, fn_name, args, expander, stack):
+def revisionid_fn(ctx, fn_name, args, expander):
+    """Implements the REVISIONID magic word."""
+    # We just return a dash, similar to "miser mode" in MediaWiki."""
+    return "-"
+
+
+def revisionuser_fn(ctx, fn_name, args, expander):
+    """Implements the REVISIONUSER magic word."""
+    # We always return AnonymousUser
+    return "AnonymousUser"
+
+
+def displaytitle_fn(ctx, fn_name, args, expander):
     """Implements the DISPLAYTITLE magic word/parser function."""
     t = expander(args[0]) if args else ""
     # XXX this should at least remove html tags h1 h2 h3 h4 h5 h6 div blockquote
     # ol ul li hr table tr th td dl dd caption p ruby rb rt rtc rp br
     # Looks as if this should also set the display title for the page in ctx???
-    return t
+    # XXX I think this parser function exists for the side effect of
+    # setting page title
+    return ""
 
-def defaultsort_fn(ctx, fn_nae, args, expander, stack):
+def defaultsort_fn(ctx, fn_nae, args, expander):
     """Implements the DEFAULTSORT magic word/parser function."""
     # XXX apparently this should set the title by which this page is
     # sorted in category listings
     return ""
 
-def lc_fn(ctx, fn_name, args, expander, stack):
+def lc_fn(ctx, fn_name, args, expander):
     """Implements the lc parser function (lowercase)."""
     return expander(args[0]).strip().lower() if args else ""
 
 
-def lcfirst_fn(ctx, fn_name, args, expander, stack):
+def lcfirst_fn(ctx, fn_name, args, expander):
     """Implements the lcfirst parser function (lowercase first character)."""
     t = expander(args[0]).strip() if args else ""
     if not t:
@@ -276,28 +438,55 @@ def lcfirst_fn(ctx, fn_name, args, expander, stack):
     return t[0].lower() + t[1:]
 
 
-def uc_fn(ctx, fn_name, args, expander, stack):
+def uc_fn(ctx, fn_name, args, expander):
     """Implements the uc parser function (uppercase)."""
     t = expander(args[0]).strip() if args else ""
     return t.upper()
 
 
-def ucfirst_fn(ctx, fn_name, args, expander, stack):
+def ucfirst_fn(ctx, fn_name, args, expander):
     """Implements the ucfirst parser function (capitalize first character)."""
     t = expander(args[0]).strip() if args else ""
     return capitalizeFirstOnly(t)
 
 
-def dateformat_fn(ctx, fn_name, args, expander, stack):
+def formatnum_fn(ctx, fn_name, args, expander):
+    """Implements the formatnum parser function."""
+    arg0 = expander(args[0]).strip() if args else ""
+    arg1 = expander(args[1]).strip() if len(args) >= 2 else ""
+    if arg1 == "R":
+        # Reverse formatting
+        # XXX this is a very simplified implementation, should handle more cases
+        return re.sub(r",", "", arg0)
+    if arg1 == "NOSEP":
+        sep = ""
+    else:
+        sep = ","
+    comma = "."  # Really should depend on locale
+    # XXX implement support for non-english locales for digits
+    orig = arg0.split(".")
+    first = orig[0]
+    parts = []
+    first = "".join(reversed(first))
+    for i in range(0, len(first), 3):
+        parts.append("".join(reversed(first[i: i + 3])))
+    parts = [sep.join(reversed(parts))]
+    if len(orig) > 1:
+        parts.append(comma)
+        parts.append(".".join(orig[1:]))
+    return "".join(parts)
+
+
+def dateformat_fn(ctx, fn_name, args, expander):
     """Implements the #dateformat (= #formatdate) parser function."""
     arg0 = expander(args[0]) if args else ""
-    if not re.search(r"\d\d\d", arg0):
-        arg0 += " 3333"
-    dt = dateparser.parse(arg0)
+    arg0x = arg0
+    if not re.search(r"\d\d\d", arg0x):
+        arg0x += " 3333"
+    dt = dateparser.parse(arg0x)
     if not dt:
-        print("{}: invalid date format in {}: {!r}"
-              "".format(ctx, fn_name, arg0))
-        dt = datetime.datetime.utcnow()
+        # It seems this should return invalid dates as-is
+        return arg0
     fmt = expander(args[1]) if len(args) > 1 else "ISO 8601"
     # This is supposed to format according to user preferences by default.
     if fmt in ("ISO 8601", "ISO8601") and dt.year == 0:
@@ -327,11 +516,25 @@ def dateformat_fn(ctx, fn_name, args, expander, stack):
     return dt.isoformat()
 
 
-def fullurl_fn(ctx, fn_name, args, expander, stack):
+def localurl_fn(ctx, fn_name, args, expander):
+    """Implements the localurl parser function."""
+    arg0 = expander(args[0]).strip() if args else ctx.title
+    arg1 = expander(args[1]).strip() if len(args) >= 2 else ""
+    # XXX handle interwiki prefixes in arg0
+    if arg1:
+        url = "/w/index.php?title={}&{}".format(
+            urllib.parse.quote_plus(arg0), arg1)
+    else:
+        url = "/wiki/{}".format(wikiurlencode(arg0))
+    return url
+
+
+def fullurl_fn(ctx, fn_name, args, expander):
     """Implements the fullurl parser function."""
     arg0 = expander(args[0]).strip() if args else ""
     # XXX handle interwiki prefixes in arg0
-    url = "//dummy.host/index.php?title=" + urllib.parse.quote_plus(arg0)
+    url = "//{}/index.php?title={}".format(
+        SERVER_NAME, urllib.parse.quote_plus(arg0))
     if len(args) > 1:
         for arg in args[1:]:
             arg = expander(arg).strip()
@@ -344,7 +547,7 @@ def fullurl_fn(ctx, fn_name, args, expander, stack):
     return url
 
 
-def urlencode_fn(ctx, fn_name, args, expander, stack):
+def urlencode_fn(ctx, fn_name, args, expander):
     """Implements the urlencode parser function."""
     arg0 = expander(args[0]) if args else ""
     fmt = expander(args[1]) if len(args) > 1 else "QUERY"
@@ -354,11 +557,16 @@ def urlencode_fn(ctx, fn_name, args, expander, stack):
     elif fmt == "QUERY":
         return urllib.parse.quote_plus(url)
     # All else in WIKI encoding
+    return wikiurlencode(url)
+
+
+def wikiurlencode(url):
+    assert isinstance(url, str)
     url = re.sub(r"\s+", "_", url)
     return urllib.parse.quote(url, safe="/:")
 
 
-def anchorencode_fn(ctx, fn_name, args, expander, stack):
+def anchorencode_fn(ctx, fn_name, args, expander):
     """Implements the urlencode parser function."""
     anchor = expander(args[0]).strip() if args else ""
     anchor = re.sub(r"\s+", "_", anchor)
@@ -483,7 +691,7 @@ add_ns(namespaces, module_ns)
 add_ns(namespaces, module_talk_ns)
 
 
-def ns_fn(ctx, fn_name, args, expander, stack):
+def ns_fn(ctx, fn_name, args, expander):
     """Implements the ns parser function."""
     t = expander(args[0]).strip().upper() if args else ""
     if t and t.isdigit():
@@ -491,7 +699,7 @@ def ns_fn(ctx, fn_name, args, expander, stack):
         ns = namespaces.get(t)
     else:
         for ns in namespaces.values():
-            print("checking", ns.name)
+            # print("checking", ns.name)
             if ns.name and t == ns.name.upper():
                 break
             if ns.canonicalName and t == ns.canonicalName.upper():
@@ -509,7 +717,7 @@ def ns_fn(ctx, fn_name, args, expander, stack):
     return ns.name
 
 
-def titleparts_fn(ctx, fn_name, args, expander, stack):
+def titleparts_fn(ctx, fn_name, args, expander):
     """Implements the #titleparts parser function."""
     t = expander(args[0]).strip() if args else ""
     arg1 = expander(args[1]).strip() if len(args) >= 2 else ""
@@ -615,7 +823,7 @@ binary_or_fns = {
     "or": lambda x, y: 1 if x or y else 0,
 }
 
-def expr_fn(ctx, fn_name, args, expander, stack):
+def expr_fn(ctx, fn_name, args, expander):
     """Implements the #titleparts parser function."""
     full_expr = expander(args[0]).strip().lower() if args else ""
     full_expr = full_expr or ""
@@ -626,10 +834,11 @@ def expr_fn(ctx, fn_name, args, expander, stack):
 
     def expr_error(tok):
         if tok is None:
-            tok = "<end>"
-        ctx.warning("{}: #expr error near {} in {!r} at {}"
-                    .format(ctx, tok, full_expr, stack))
-        return ""
+            tok = "&lt;end&gt;"
+        #ctx.warning("#expr error near {} in {!r}"
+        #            .format(tok, full_expr))
+        return ('<strong class="error">Expression error near {}</strong>'
+                .format(tok))
 
     def get_token():
         nonlocal tokidx
@@ -670,8 +879,7 @@ def expr_fn(ctx, fn_name, args, expander, stack):
             return math.e
         if tok == "pi":
             return math.pi
-        if tok == "nil":
-            #print("{}: nil in expr {!r} at {}".format(ctx, full_expr, stack))
+        if tok == ".":
             return 0
         return expr_error(tok)
 
@@ -755,13 +963,13 @@ def expr_fn(ctx, fn_name, args, expander, stack):
     return str(ret)
 
 
-def padleft_fn(ctx, fn_name, args, expander, stack):
+def padleft_fn(ctx, fn_name, args, expander):
     """Implements the padleft parser function."""
     v = expander(args[0]) if args else ""
     cnt = expander(args[1]).strip() if len(args) >= 2 else "0"
     pad = expander(args[2]) if len(args) >= 3 and args[2] else "0"
     if not cnt.isdigit():
-        print("{}: pad length is not integer: {!r}".format(ctx, cnt))
+        ctx.warning("pad length is not integer: {!r}".format(cnt))
         cnt = 0
     else:
         cnt = int(cnt)
@@ -772,14 +980,14 @@ def padleft_fn(ctx, fn_name, args, expander, stack):
     return v
 
 
-def padright_fn(ctx, fn_name, args, expander, stack):
+def padright_fn(ctx, fn_name, args, expander):
     """Implements the padright parser function."""
     v = expander(args[0]) if args else ""
     cnt = expander(args[1]).strip() if len(args) >= 2 else "0"
     arg2 = expander(args[2]) if len(args) >= 3 and args[2] else "0"
     pad = arg2 if len(args) >= 3 and arg2 else "0"
     if not cnt.isdigit():
-        print("{}: pad length is not integer: {!r}".format(ctx, cnt))
+        ctx.warning("pad length is not integer: {!r}".format(cnt))
         cnt = 0
     else:
         cnt = int(cnt)
@@ -790,10 +998,10 @@ def padright_fn(ctx, fn_name, args, expander, stack):
     return v
 
 
-def plural_fn(ctx, fn_name, args, expander, stack):
+def plural_fn(ctx, fn_name, args, expander):
     """Implements the #plural parser function."""
     expr = expander(args[0]).strip() if args else "0"
-    v = expr_fn(ctx, fn_name, [expr], lambda x: x, stack)
+    v = expr_fn(ctx, fn_name, [expr], lambda x: x)
     # XXX for some language codes, this is more complex.  See {{plural:...}} in
     # https://www.mediawiki.org/wiki/Help:Magic_words
     if v == 1:
@@ -801,13 +1009,121 @@ def plural_fn(ctx, fn_name, args, expander, stack):
     return expander(args[2]).strip() if len(args) >= 3 else ""
 
 
-def len_fn(ctx, fn_name, args, expander, stack):
+def month_num_days(ctx, t):
+    mdays = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
+    v = mdays[t.month - 1]
+    if t.month == 2:
+        if t.year % 4 == 0 and (t.year % 100 != 0 or t.year % 400 == 0):
+            v = 29
+    return v
+
+
+time_fmt_map = {
+    "Y": "%Y",
+    "y": "%y",
+    "L": lambda ctx, t: 1 if (t.year % 4 == 0 and
+                              (t.year % 100 != 0 or
+                               t.year % 400 == 0)) else 0,
+    "o": "%G",
+    "n": lambda ctx, t: t.month,
+    "m": "%m",
+    "M": "%b",
+    "F": "%B",
+    "xg": "%B",  # Should be in genitive
+    "j": lambda ctx, t: t.day,
+    "d": "%d",
+    "z": lambda ctx, t:
+        (t - datetime.datetime(year=t.year, month=1, day=1,
+                               tzinfo=t.tzinfo)).days,
+    "W": "%V",
+    "N": "%u",
+    "w": "%w",
+    "D": "%a",
+    "l": "%A",
+    "a": "%p",  # Should be lowercase
+    "A": "%p",  # Should be uppercase
+    "g": lambda ctx, t: t.hour % 12,
+    "h": "%I",
+    "G": lambda ctx, t: t.hour,
+    "H": "%H",
+    "i": "%M",
+    "s": "%S",
+    "U": lambda ctx, t: int(t.timestamp()),
+    "e": "%Z",
+    "I": lambda ctx, t: "1" if t.dst() and t.dst().seconds != 0 else "0",
+    "0": lambda ctx, t: t.strftime("%z")[:5],
+    "P": lambda ctx, t: t.strftime("%z")[:3] + ":" + t.strftime("%z")[3:5],
+    "T": "%Z",
+    "Z": lambda ctx, t: 0 if t.utcoffset() == None else t.utcoffset().seconds,
+    "t": month_num_days,
+    "c": lambda ctx, t: t.isoformat(),
+    "r": lambda ctx, t: t.strftime("%a, %d %b %Y %H:%M:%S {}").format(
+        t.strftime("%z")[:5]),
+    # XXX non-gregorian calendar values
+}
+
+
+def time_fn(ctx, fn_name, args, expander):
+    """Implements the #time parser function."""
+    fmt = expander(args[0]).strip() if args else ""
+    dt = expander(args[1]).strip() if len(args) >= 2 else ""
+    lang = expander(args[2]).strip() if len(args) >= 3 else "en"
+    loc = expander(args[3]).strip() if len(args) >= 4 else ""
+
+    dt = re.sub(r"\+", " in ", dt)
+    if not dt:
+        dt = "now"
+
+    settings = { "RETURN_AS_TIMEZONE_AWARE": True}
+    if loc in ("", "0"):
+        dt += " UTC"
+
+    if dt.startswith("@"):
+        try:
+            t = datetime.datetime.fromtimestamp(float(dt[1:]))
+        except ValueError:
+            ctx.warning("bad time syntax in {}: {!r}"
+                        .format(fn_name, dt))
+            return ('<strong class="error">Bad time syntax: {}</strong>'
+                    .format(html.escape(dt)))
+    else:
+        t = dateparser.parse(dt, settings=settings)
+        if t is None:
+            ctx.warning("unrecognized time syntax in {}: {!r}"
+                        .format(fn_name, dt))
+            return ('<strong class="error">Bad time syntax: {}</strong>'
+                    .format(html.escape(dt)))
+
+    # XXX looks like we should not adjust the time
+    #if t.utcoffset():
+    #    t -= t.utcoffset()
+
+    def fmt_repl(m):
+        f = m.group(0)
+        if len(f) > 1 and f.startswith('"') and f.endswith('"'):
+            return f[1:-1]
+        if f in time_fmt_map:
+            v = time_fmt_map[f]
+            if isinstance(v, str):
+                return v
+            assert callable(v)
+            v = v(ctx, t)
+            if not isinstance(v, str):
+                v = str(v)
+            return v
+        return f
+
+    fmt = re.sub(r'(x[mijkot]?)?[^"]|"[^"]*"', fmt_repl, fmt)
+    return t.strftime(fmt)
+
+
+def len_fn(ctx, fn_name, args, expander):
     """Implements the #len parser function."""
     v = expander(args[0]).strip() if args else ""
     return str(len(v))
 
 
-def pos_fn(ctx, fn_name, args, expander, stack):
+def pos_fn(ctx, fn_name, args, expander):
     """Implements the #pos parser function."""
     arg0 = expander(args[0]).strip() if args else ""
     arg1 = expander(args[1]) or " " if len(args) >= 2 else " "
@@ -821,7 +1137,7 @@ def pos_fn(ctx, fn_name, args, expander, stack):
     return ""
 
 
-def rpos_fn(ctx, fn_name, args, expander, stack):
+def rpos_fn(ctx, fn_name, args, expander):
     """Implements the #rpos parser function."""
     arg0 = expander(args[0]).strip() if args else ""
     arg1 = expander(args[1]) or " " if len(args) >= 2 else " "
@@ -835,7 +1151,7 @@ def rpos_fn(ctx, fn_name, args, expander, stack):
     return "-1"
 
 
-def sub_fn(ctx, fn_name, args, expander, stack):
+def sub_fn(ctx, fn_name, args, expander):
     """Implements the #sub parser function."""
     arg0 = expander(args[0]).strip() if args else ""
     start = expander(args[1]).strip() if len(args) >= 2 else ""
@@ -858,14 +1174,14 @@ def sub_fn(ctx, fn_name, args, expander, stack):
     return arg0[start : start + length]
 
 
-def pad_fn(ctx, fn_name, args, expander, stack):
+def pad_fn(ctx, fn_name, args, expander):
     """Implements the pad parser function."""
     v = expander(args[0]).strip() if args else ""
     cnt = expander(args[1]).strip() if len(args) >= 2 else ""
     pad = expander(args[2]) if len(args) >= 3 and args[2] else "0"
     direction = expander(args[3]) if len(args) >= 4 else ""
     if not cnt.isdigit():
-        print("{}: pad length is not integer: {!r}".format(ctx, cnt))
+        ctx.warning("pad length is not integer: {!r}".format(cnt))
         cnt = 0
     else:
         cnt = int(cnt)
@@ -882,7 +1198,7 @@ def pad_fn(ctx, fn_name, args, expander, stack):
     return v
 
 
-def replace_fn(ctx, fn_name, args, expander, stack):
+def replace_fn(ctx, fn_name, args, expander):
     """Implements the #replace parser function."""
     arg0 = expander(args[0]).strip() if args else ""
     arg1 = expander(args[1]) or " " if len(args) >= 2 else " "
@@ -890,7 +1206,7 @@ def replace_fn(ctx, fn_name, args, expander, stack):
     return arg0.replace(arg1, arg2)
 
 
-def explode_fn(ctx, fn_name, args, expander, stack):
+def explode_fn(ctx, fn_name, args, expander):
     """Implements the #explode parser function."""
     arg0 = expander(args[0]).strip() if args else ""
     delim = expander(args[1]) or " " if len(args) >= 2 else " "
@@ -907,7 +1223,6 @@ def explode_fn(ctx, fn_name, args, expander, stack):
     parts = arg0.split(delim)
     if limit > 0 and len(parts) > limit:
         parts = parts[:limit - 1] + [delim.join(parts[limit - 1:])]
-    print("parts", parts)
     if position < 0:
         position = len(parts) + position
     if position < 0 or position >= len(parts):
@@ -915,15 +1230,15 @@ def explode_fn(ctx, fn_name, args, expander, stack):
     return parts[position]
 
 
-def urldecode_fn(ctx, fn_name, args, expander, stack):
+def urldecode_fn(ctx, fn_name, args, expander):
     """Implements the #urldecode parser function."""
     arg0 = expander(args[0]).strip() if args else ""
     ret = urllib.parse.unquote_plus(arg0)
     return ret
 
 
-def unimplemented_fn(ctx, fn_name, args, expander, stack):
-    print("{}: unimplemented parserfn {} at {}".format(ctx, fn_name, stack))
+def unimplemented_fn(ctx, fn_name, args, expander):
+    ctx.error("unimplemented parserfn {}".format(fn_name))
     return "{{" + fn_name + ":" + "|".join(map(str, args)) + "}}"
 
 
@@ -935,20 +1250,20 @@ PARSER_FUNCTIONS = {
     "FULLPAGENAME": fullpagename_fn,
     "PAGENAME": pagename_fn,
     "BASEPAGENAME": basepagename_fn,
-    "ROOTPAGENAME": unimplemented_fn,
+    "ROOTPAGENAME": rootpagename_fn,
     "SUBPAGENAME": subpagename_fn,
     "ARTICLEPAGENAME": unimplemented_fn,
     "SUBJECTPAGENAME": unimplemented_fn,
-    "TALKPAGENAME": unimplemented_fn,
+    "TALKPAGENAME": talkpagename_fn,
     "NAMESPACENUMBER": namespacenumber_fn,
     "NAMESPACE": namespace_fn,
     "ARTICLESPACE": unimplemented_fn,
-    "SUBJECTSPACE": unimplemented_fn,
-    "TALKSPACE": unimplemented_fn,
-    "FULLPAGENAMEE": unimplemented_fn,
-    "PAGENAMEE": unimplemented_fn,
+    "SUBJECTSPACE": subjectspace_fn,
+    "TALKSPACE": talkspace_fn,
+    "FULLPAGENAMEE": fullpagenamee_fn,
+    "PAGENAMEE": pagenamee_fn,
     "BASEPAGENAMEE": unimplemented_fn,
-    "ROOTPAGENAMEE": unimplemented_fn,
+    "ROOTPAGENAMEE": rootpagenamee_fn,
     "SUBPAGENAMEE": unimplemented_fn,
     "ARTICLEPAGENAMEE": unimplemented_fn,
     "SUBJECTPAGENAMEE": unimplemented_fn,
@@ -960,15 +1275,15 @@ PARSER_FUNCTIONS = {
     "TALKSPACEE": unimplemented_fn,
     "SHORTDESC": unimplemented_fn,
     "SITENAME": unimplemented_fn,
-    "SERVER": unimplemented_fn,
-    "SERVERNAME": unimplemented_fn,
+    "SERVER": server_fn,
+    "SERVERNAME": servername_fn,
     "SCRIPTPATH": unimplemented_fn,
     "CURRENTVERSION": unimplemented_fn,
     "CURRENTYEAR": currentyear_fn,
     "CURRENTMONTH": currentmonth_fn,
     "CURRENTMONTH1": currentmonth1_fn,
-    "CURRENTMONTHNAME": unimplemented_fn,
-    "CURRENTMONTHABBREV": unimplemented_fn,
+    "CURRENTMONTHNAME": currentmonthname_fn,
+    "CURRENTMONTHABBREV": currentmonthabbrev_fn,
     "CURRENTDAY": currentday_fn,
     "CURRENTDAY2": currentday2_fn,
     "CUEEWNTDOW": currentdow_fn,
@@ -989,12 +1304,13 @@ PARSER_FUNCTIONS = {
     "LOCALHOUR": unimplemented_fn,
     "LOCALWEEK": unimplemented_fn,
     "LOCALTIMESTAMP": unimplemented_fn,
+    "REVISIONID": revisionid_fn,
     "REVISIONDAY": unimplemented_fn,
     "REVISIONDAY2": unimplemented_fn,
     "REVISIONMONTH": unimplemented_fn,
     "REVISIONYEAR": unimplemented_fn,
     "REVISIONTIMESTAMP": unimplemented_fn,
-    "REVISIONUSER": unimplemented_fn,
+    "REVISIONUSER": revisionuser_fn,
     "NUMBEROFPAGES": unimplemented_fn,
     "NUMBEROFARTICLES": unimplemented_fn,
     "NUMBEROFFILES": unimplemented_fn,
@@ -1016,17 +1332,17 @@ PARSER_FUNCTIONS = {
     "lcfirst": lcfirst_fn,
     "uc": uc_fn,
     "ucfirst": ucfirst_fn,
-    "formatnum": unimplemented_fn,
+    "formatnum": formatnum_fn,
     "#dateformat": dateformat_fn,
     "#formatdate": dateformat_fn,
     "padleft": padleft_fn,
     "padright": padright_fn,
     "plural": plural_fn,
-    "#time": unimplemented_fn,
+    "#time": time_fn,
     "#timel": unimplemented_fn,
     "gender": unimplemented_fn,
     "#tag": tag_fn,
-    "localurl": unimplemented_fn,
+    "localurl": localurl_fn,
     "fullurl": fullurl_fn,
     "canonicalurl": unimplemented_fn,
     "filepath": unimplemented_fn,
@@ -1039,12 +1355,12 @@ PARSER_FUNCTIONS = {
     "#expr": expr_fn,
     "#if": if_fn,
     "#ifeq": ifeq_fn,
-    "#iferror": unimplemented_fn,
+    "#iferror": iferror_fn,
     "#ifexpr": ifexpr_fn,
     "#ifexist": ifexist_fn,
     "#switch": switch_fn,
     "#babel": unimplemented_fn,
-    "#categorytree": unimplemented_fn,
+    "#categorytree": (categorytree_fn, True),  # This takes kwargs
     "#coordinates": unimplemented_fn,
     "#invoke": unimplemented_fn,
     "#language": unimplemented_fn,
@@ -1078,16 +1394,23 @@ PARSER_FUNCTIONS = {
 }
 
 
-def call_parser_function(ctx, fn_name, args, expander, stack):
+def call_parser_function(ctx, fn_name, args, expander):
     """Calls the given parser function with the given arguments."""
     assert isinstance(fn_name, str)
     assert isinstance(args, (list, tuple, dict))
     assert callable(expander)
-    assert isinstance(stack, list)
+    if fn_name not in PARSER_FUNCTIONS:
+        ctx.error("unrecognized parser function {!r}".format(fn_name))
+        return ""
     fn = PARSER_FUNCTIONS[fn_name]
+    accept_keyed_args = False
+    if isinstance(fn, tuple):
+        accept_keyed_args = fn[1]
+        fn = fn[0]
+    assert callable(fn)
     have_keyed_args = False
-    if isinstance(args, dict):
-        dict_args = args.copy()
+    if isinstance(args, dict) and not accept_keyed_args:
+        # Convert from dict to vector, no keyed args allowed
         new_args = []
         for i in range(1, 1000):
             v = args.get(i, None)
@@ -1098,10 +1421,25 @@ def call_parser_function(ctx, fn_name, args, expander, stack):
             del args[i]
         have_keyed_args = len(args) > 0
         args = new_args
-    else:
-        dict_args = dict(zip(range(1, len(args) + 1), args))
-    if have_keyed_args:
-        print("{}: {}: ERROR: named arguments not supported: {} at {}"
-              .format(ctx.title, fn_name, args, stack))
+    elif accept_keyed_args:
+        # Convert from vector to keyed args
+        ht = {}
+        i = 1
+        for arg in args:
+            arg = str(arg)
+            ofs = arg.find("=")
+            if ofs >= 0:
+                k = arg[:ofs]
+                if k.isdigit():
+                    k = int(k)
+                arg = arg[ofs + 1:]
+            else:
+                k = i
+            ht[k] = arg
+        args = ht
+    if have_keyed_args and not accept_keyed_args:
+        ctx.error("parser function {} does not (yet) support named "
+                  "arguments: {}"
+                  .format(fn_name, args))
         return ""
-    return fn(ctx, fn_name, args, expander, stack)
+    return fn(ctx, fn_name, args, expander)
